from collections import deque

from .constants import Constants
from .message import Message


class MessageBuffer:
    """Maintains a limited history of messages grouped by round."""

    def __init__(self) -> None:
        self._history: deque[int] = deque(maxlen=Constants.MESSAGE_HISTORY_LIMIT)
        self._round_map: dict[int, list[Message]] = {}
        self._pending: list[Message] = []

    def add_message(self, message: Message) -> None:
        """
        Add a message to the pending buffer for next round.

        Args:
            message (Message): The message to store.

        """
        self._pending.append(message)

    def _rotate_to(self, new_round: int) -> None:
        """
        Prepare the buffer to store messages for a new round.

        Any messages in the pending queue (received during the previous round)
        are committed to the last round's history before starting the new one.
        This ensures that messages are only visible one round after they are
        received.

        If the maximum history size is reached, the oldest round's messages
        are discarded.

        Args:
            new_round (int): The new round to initialize in the buffer.

        """
        if (
            self._history.maxlen is not None
            and len(self._history) == self._history.maxlen
        ):
            oldest = self._history.popleft()
            del self._round_map[oldest]
        self._history.append(new_round)
        self._round_map[new_round] = list(self._pending)
        self._pending.clear()

    def get_all_messages(self) -> list[Message]:
        """
        Retrieve all messages from the buffer.

        Messages are returned in reverse-chronological order: newest round's
        messages first, oldest last.

        Returns:
            list[Message]: All stored messages, sorted by most recent round.

        """
        result: list[Message] = []
        for r in reversed(self._history):
            result.extend(self._round_map[r])
        return result

    def get_messages(self, round_num: int) -> list[Message]:
        """
        Retrieve messages from a specific round.

        Args:
            round_num (int): The round to fetch messages for.

        Returns:
            list[Message]: A copy of the messages from the round,
            or an empty list if not stored.

        """
        return list(self._round_map.get(round_num, []))

    def next_round(self, round_num: int) -> None:
        """
        Start a new round.

        Args:
            round_num (int): The round number to begin.

        """
        self._rotate_to(round_num)



from pathlib import Path
from typing import TYPE_CHECKING, cast

import numpy as np
import tensorflow as tf

from aegis_game.stub import *

if TYPE_CHECKING:
    from numpy.typing import NDArray


def think() -> None:
    """Do not remove this function, it must always be defined."""
    log("Thinking")

    # On the first round, send a request for surrounding information
    # by moving to the center (not moving). This will help initiate pathfinding.
    if get_round_number() == 1:
        move(Direction.CENTER)
        return

    # Fetch the cell at the agent's current location.
    # If you want to check a different location, use `on_map(loc)` first
    # to ensure it's within the world bounds. The agent's own location is always valid.
    cell: CellInfo = get_cell_info_at(get_location())

    # If there is a pending prediction from a save survivor for our team, predict!
    prediction_info = read_pending_predictions()
    if prediction_info:
        # grab just the first pending prediction
        surv_saved_id, image_to_predict, _ = prediction_info[0]

        # Get the path to the model file in the agent's directory
        model_path = Path("agents/agent_prediction/trained_net.keras")
        if not model_path.exists():
            log("Model not found, skipping prediction")
            return

        # Load the model and make prediction like in the example
        model = tf.keras.models.load_model(str(model_path))
        # Reshape the image for the model (similar to the example)
        image = np.reshape(image_to_predict, (28, 28))
        image = np.reshape(image, (1, 28, 28))

        # Make prediction
        predictions = model(image)
        predicted_label = cast(np.int32, np.argmax(predictions[0]))
        confidence = float(np.max(predictions[0]))

        log(f"Predicted symbol: {predicted_label} with confidence: {confidence:.4f}")

        # Send the prediction using the survivor ID from the save result
        predict(surv_saved_id, predicted_label)
    else:
        log("No pending predictions")

    # Get the top layer at the agent's current location.
    # If a survivor is present, save it and make a prediction.
    if isinstance(cell.top_layer, Survivor):
        # Save the survivor
        save()

        # After saving, we'll get a SaveResult with the image to predict
        # The prediction will be handled in the handle_save function
        return

    # Default action: Move the agent north if no other specific conditions are met.
    move(Direction.NORTH)



from aegis_game.stub import *


def think() -> None:
    """Do not remove this function, it must always be defined."""
    log("Thinking")

    # On the first round, send a request for surrounding information
    # by moving to the center (not moving). This will help initiate pathfinding.
    if get_round_number() == 1:
        move(Direction.CENTER)
        return

    # Fetch the cell at the agent's current location.
    # If you want to check a different location, use `on_map(loc)` first
    # to ensure it's within the world bounds. The agent's own location is always valid.
    cell = get_cell_info_at(get_location())

    # Get the top layer at the agent's current location.
    # If a survivor is present, save it and end the turn.
    if isinstance(cell.top_layer, Survivor):
        save()
        return

    # Default action: Move the agent north if no other specific conditions are met.
    move(Direction.NORTH)



from . import location_pb2 as _location_pb2
from . import world_object_pb2 as _world_object_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CellType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NORMAL: _ClassVar[CellType]
    SPAWN: _ClassVar[CellType]
    KILLER: _ClassVar[CellType]
    CHARGING: _ClassVar[CellType]
NORMAL: CellType
SPAWN: CellType
KILLER: CellType
CHARGING: CellType

class Cell(_message.Message):
    __slots__ = ("loc", "moveCost", "type", "agents", "layers")
    LOC_FIELD_NUMBER: _ClassVar[int]
    MOVECOST_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    AGENTS_FIELD_NUMBER: _ClassVar[int]
    LAYERS_FIELD_NUMBER: _ClassVar[int]
    loc: _location_pb2.Location
    moveCost: int
    type: CellType
    agents: _containers.RepeatedScalarFieldContainer[int]
    layers: _containers.RepeatedCompositeFieldContainer[_world_object_pb2.WorldObject]
    def __init__(self, loc: _Optional[_Union[_location_pb2.Location, _Mapping]] = ..., moveCost: _Optional[int] = ..., type: _Optional[_Union[CellType, str]] = ..., agents: _Optional[_Iterable[int]] = ..., layers: _Optional[_Iterable[_Union[_world_object_pb2.WorldObject, _Mapping]]] = ...) -> None: ...



"""
Autogenerated from API functions.

Do not modify manually.
"""
# ruff: noqa: F401
# pyright: reportReturnType=false
# pyright: reportUnusedImport=false
# pyright: reportUnusedParameter=false

from . import (
    AgentType,
    CellInfo,
    Direction,
    Location,
    Rubble,
    Survivor,
    Team,
)


def get_round_number() -> int:
    """Return the current round number."""


def get_id() -> int:
    """Return the id of the agent."""


def get_type() -> AgentType:
    """Return the type of the agent."""


def get_team() -> Team:
    """Return the current team of the agent."""


def get_location() -> Location:
    """Return the current location of the agent."""


def get_energy_level() -> int:
    """Return the current energy level of the agent."""


def move(direction: Direction) -> None:
    """
    Move the agent in the specified direction.

    Args:
        direction: The direction in which the agent should move.

    Raises:
        AgentError: If the move is invalid.

    """


def save() -> None:
    """
    Save a survivor located at the agent's current location.

    If no survivor is present, the function has no effect.

    Raises:
        AgentError: If saving is invalid according to game rules.

    """


def recharge() -> None:
    """
    Recharge the agent's energy if on a charging cell.

    Energy restored is equal to `Constants.NORMAL_CHARGE` per recharge,
    but cannot exceed `Constants.MAX_ENERGY_LEVEL`.

    Does nothing if the agent is not on a charging cell.

    """


def dig() -> None:
    """
    Dig rubble at the agent's current location.

    Raises:
        AgentError: If digging is invalid according to game rules.

    """


def drone_scan(loc: Location) -> None:
    """
    Scan a location using a drone.

    Args:
        loc: The location to scan.

    Raises:
        AgentError: If drone scan is not enabled or location is invalid.

    """


def get_cell_info_at(loc: Location) -> CellInfo:
    """
    Return the cell info at a given location.

    If the location is adjacent (1 tile away) to the agent or has been scanned by a drone,
    all layers and agents at that location are visible. Otherwise, only the top layer is
    visible and agent presence is hidden.

    If `HIDDEN_MOVE_COSTS` feature is enabled, unvisited cells have `move_cost = 1`.

    Args:
        loc: The location to query.

    Returns:
        Information about the cell at the given location.

    """


def log(*args: object) -> None:
    """
    Log a message.

    Args:
        *args: One or more items to log.

    """


def on_map(loc: Location) -> bool:
    """
    Check whether a location is within the bounds of the world.

    Args:
        loc: The location to check.

    Returns:
        True if the location is on the map, False otherwise.

    """


def get_survs() -> list[Location]:
    """Return a list of survivor locations."""


def get_charging_cells() -> list[Location]:
    """Return a list of charging locations."""



import numpy as np
import os
import matplotlib.pyplot as plt

values = np.load(os.path.join("sys_files", "values.npy"))
labels = np.load(os.path.join("sys_files", "labels.npy"))

# shuffle data
data_size = values.shape[0]
indices = np.arange(data_size)
np.random.shuffle(indices)

values = values[indices]
labels = labels[indices]

# split data
split_index = int(0.8 * data_size)
x_train, x_test = values[:split_index], values[split_index:]
y_train, y_test = labels[:split_index], labels[split_index:]

plt.imshow(values[0], cmap="gray")
plt.title(f"Label: {labels[0]}")
plt.axis("off")
plt.show()

# save data to the 4 spots ! !
agent_training_output_dir = os.path.join("src", "assist", "model_training_data")
ares_testing_output_dir = os.path.join(
    "src", "aegis", "agent_predictions", "model_testing_data"
)

os.makedirs(agent_training_output_dir, exist_ok=True)
os.makedirs(ares_testing_output_dir, exist_ok=True)

np.save(os.path.join(agent_training_output_dir, "x_train_a3.npy"), x_train)
np.save(os.path.join(agent_training_output_dir, "y_train_a3.npy"), y_train)
np.save(os.path.join(ares_testing_output_dir, "x_test_a3.npy"), x_test)
np.save(os.path.join(ares_testing_output_dir, "y_test_a3.npy"), y_test)



class Constants:
    # Charging constants
    NORMAL_CHARGE: int = 5
    LOW_CHARGE: int = 1
    SUPER_CHARGE: int = 20
    MAX_ENERGY_LEVEL: int = 1000

    # World constants
    WORLD_MIN: int = 3
    WORLD_MAX: int = 30

    # Game constants
    DEFAULT_MAX_ROUNDS: int = 1000
    MESSAGE_HISTORY_LIMIT: int = 5
    MAX_TURN_TIME_LIMIT: float = 1.0
    INITIAL_TEAM_LUMENS: int = 100

    # Points constants
    SURVIVOR_SAVE_ALIVE_SCORE: int = 100  # Only used if SURV_HEALTH_DECAY_RATE is 0
    SURVIVOR_SAVE_DEAD_SCORE: int = 50  # Only used if SURV_HEALTH_DECAY_RATE is 0
    MIN_SURVIVOR_SAVE_SCORE: int = 10  # Only used if SURV_HEALTH_DECAY_RATE is 0, defines lowest possible score for a survivor save
    PRED_CORRECT_SCORE: int = 10
    ALIVE_AGENT_SCORE: int = 50

    # Cooldown constants
    COOLDOWN_LIMIT: int = 10
    COOLDOWN_TICK: int = 10

    # Energy constants
    OBSERVE_ENERGY_COST: int = 1
    SAVE_ENERGY_COST: int = 1
    PREDICTION_ENERGY_COST: int = 1
    MOVE_ENERGY_COST: int = 1
    DRONE_SCAN_ENERGY_COST: int = 2
    ENERGY_PENALTY_FOR_ERRORS: int = -10

    # Symbol prediction constants
    NUM_OF_TESTING_IMAGES: int = 704

    # Other
    DRONE_SCAN_DURATION: int = 5
    LUMENS_PER_SAVE: int = 20
    LUMENS_PER_ROUND: int = 5



from pathlib import Path

from google.protobuf.message import DecodeError

from .common import Cell, Location
from .common.objects import Rubble, Survivor
from .schemas import world_pb2
from .schemas.cell_pb2 import Cell as PbCell
from .schemas.cell_pb2 import CellType
from .schemas.world_object_pb2 import SurvivorState
from .world import World


def get_cell_type(cell: Cell) -> CellType:
    if cell.is_spawn_cell():
        return CellType.SPAWN
    if cell.is_charging_cell():
        return CellType.CHARGING
    if cell.is_killer_cell():
        return CellType.KILLER
    return CellType.NORMAL


def cell_type_from_proto(cell: Cell, cell_type: CellType) -> None:
    if cell_type == CellType.SPAWN:
        cell.set_spawn_cell()
    elif cell_type == CellType.CHARGING:
        cell.set_charging_cell()
    elif cell_type == CellType.KILLER:
        cell.set_killer_cell()


def cell_from_proto(proto_cell: PbCell) -> Cell:
    loc = proto_cell.loc
    cell = Cell(loc.x, loc.y)

    cell_type_from_proto(cell, proto_cell.type)
    cell.move_cost = proto_cell.moveCost
    cell.agents = list(proto_cell.agents)

    for layer_proto in proto_cell.layers:
        if layer_proto.HasField("survivor"):
            s = layer_proto.survivor
            # Convert protobuf state to Survivor.State enum
            state = Survivor.State.ALIVE if s.state == 0 else Survivor.State.DEAD
            survivor = Survivor(s.id, s.health, state)
            cell.add_layer(survivor)
        elif layer_proto.HasField("rubble"):
            r = layer_proto.rubble
            rubble = Rubble(
                r.id,
                r.energy_required,
                r.agents_required,
            )
            cell.add_layer(rubble)

    return cell


def serialize_world(world: World) -> world_pb2.World:
    proto_world = world_pb2.World()
    proto_world.width = world.width
    proto_world.height = world.height
    proto_world.seed = world.seed
    proto_world.start_energy = world.start_energy
    proto_world.total_survivors = world.total_survivors

    for cell in world.cells:
        proto_cell = proto_world.cells.add()
        proto_cell.loc.x = cell.location.x
        proto_cell.loc.y = cell.location.y
        proto_cell.moveCost = cell.move_cost
        proto_cell.type = get_cell_type(cell)
        proto_cell.agents.extend(cell.agents)

        for layer in cell.get_layers():
            layer_proto = proto_cell.layers.add()
            if isinstance(layer, Survivor):
                survivor_proto = layer_proto.survivor
                survivor_proto.id = layer.id
                survivor_proto.health = layer.health
                # Use the actual survivor state instead of inferring from health
                survivor_proto.state = (
                    SurvivorState.ALIVE if layer.is_alive() else SurvivorState.DEAD
                )
            elif isinstance(layer, Rubble):
                rubble_proto = layer_proto.rubble
                rubble_proto.id = layer.id
                rubble_proto.energy_required = layer.energy_required
                rubble_proto.agents_required = layer.agents_required

    return proto_world


def init_spawns_from_proto(world: world_pb2.World) -> dict[Location, int]:
    spawns: dict[Location, int] = {}

    for spawn in world.init_spawns:
        loc = Location(spawn.loc.x, spawn.loc.y)
        spawns[loc] = spawns.get(loc, 0) + spawn.amount

    return spawns


def deserialize_world(data: bytes) -> World:
    try:
        proto_world = world_pb2.World()
        _ = proto_world.ParseFromString(data)
    except DecodeError as e:
        error = "Failed to decode binary world file"
        raise ValueError(error) from e

    cells = [cell_from_proto(proto_cell) for proto_cell in proto_world.cells]

    spawns = init_spawns_from_proto(proto_world)
    return World(
        proto_world.width,
        proto_world.height,
        proto_world.seed,
        proto_world.start_energy,
        cells,
        spawns,
    )


def load_world(filename: Path) -> World:
    with filename.open("rb") as file:
        return deserialize_world(file.read())



# pyright: reportMissingTypeStubs = false
import base64
import queue
import threading
import time
from typing import NamedTuple

from websocket_server import WebsocketServer

from .logger import LOGGER


class Client(NamedTuple):
    """Represents a connected client."""

    id: int
    handler: object
    address: str


class WebSocketServer:
    def __init__(self, *, wait_for_client: bool) -> None:
        """Initialize a new server."""
        self._host: str = "localhost"
        self._port: int = 6003
        self._wait_for_client: bool = wait_for_client
        self._connected: bool = False
        self._done: bool = False
        self._server: WebsocketServer | None = None
        self._previous_events: list[str] = []
        self._incoming_events: queue.Queue[str] = queue.Queue()
        self._queue_thread: threading.Thread = threading.Thread(
            target=self._process_queue,
        )
        self._lock: threading.Lock = threading.Lock()

    def _process_queue(self) -> None:
        """Events to process that are in the event queue."""
        try:
            while not self._done:
                try:
                    event = self._incoming_events.get(timeout=0.3)
                    self._process_event(event)
                except queue.Empty:
                    pass

            while not self._incoming_events.empty():
                event = self._incoming_events.get()
                self._process_event(event)
        except (OSError, RuntimeError) as e:
            LOGGER.exception("Error processing queue: %s", e)

    def _process_event(self, event: str) -> None:
        if self._server is not None:
            with self._lock:
                for client in self._server.clients:  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                    self._server.send_message(client, event)  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                self._previous_events.append(event)

    def add_event(self, event: bytes) -> None:
        if self._done:
            error = "Can't add event, server already finished!"
            raise RuntimeError(error)
        encoded_event = base64.b64encode(event).decode("utf-8")
        self._incoming_events.put(encoded_event)

    def _on_open(self, client: dict[str, object], server: WebsocketServer) -> None:
        self._connected = True
        for event in self._previous_events:
            server.send_message(client, event)  # pyright: ignore[reportUnknownMemberType]

    def start(self) -> None:
        if not self._wait_for_client:
            return

        self._server = WebsocketServer(self._host, self._port)
        self._server.set_fn_new_client(self._on_open)  # pyright: ignore[reportUnknownMemberType]

        self._queue_thread.start()
        self._server.run_forever(threaded=True)

        while not self._connected:
            time.sleep(0.3)

    def shutdown_gracefully(self) -> None:
        if self._server is None:
            return
        self._server.keep_alive = False
        self._server._disconnect_clients_gracefully(1000, bytes("", encoding="utf-8"))  # pyright: ignore[reportPrivateUsage]  # noqa: SLF001
        # These bottom two are flipped from regular order in websocket_server
        self._server.shutdown()
        self._server.server_close()

    def finish(self) -> None:
        if not self._wait_for_client:
            return
        self._done = True
        try:
            self._queue_thread.join()
            self.shutdown_gracefully()
        except (OSError, RuntimeError) as e:
            LOGGER.exception("Error shutting down server: %s", e)

    def set_wait_for_client(self) -> None:
        self._wait_for_client = True



"""
Generate `aegis_game/stub.py` from API functions with decorators.

This script extracts only the necessary function stubs with their docstrings.
It filters functions based on enabled feature flags and writes the resulting stubs
into `aegis_game/stub.py`.

Designed to be run manually when updating the public-facing stub interface
based on feature flags or function changes.
"""

import ast
from dataclasses import dataclass
from importlib import resources
from pathlib import Path

from _aegis_game.aegis_config import has_feature
from _aegis_game.types.others import FeatureKey


@dataclass
class FunctionStub:
    """Represents a function stub with its metadata."""

    name: str
    signature: str
    docstring: str | None

    def format(self, *, is_last: bool = False) -> str:
        """Format the function stub as a string."""
        stub = self.signature
        if self.docstring:
            stub += self._format_docstring()
        if not is_last:
            stub += "\n"
        return stub

    def _format_docstring(self) -> str:
        """Format the docstring with proper indentation."""
        if self.docstring is None:
            return ""

        lines = self.docstring.splitlines()
        if len(lines) == 1:
            return f'\n    """{lines[0]}"""'

        indented_lines = "\n".join(
            f"    {line}" if line.strip() else "" for line in lines
        )
        return f'\n    """\n{indented_lines}\n\n    """'


class ASTFunctionParser:
    """Handles parsing of function definitions from AST nodes."""

    @staticmethod
    def format_return_annotation(returns: ast.expr | None) -> str:
        """Format the return type annotation of a function."""
        return ast.unparse(returns) if returns else ""

    @staticmethod
    def format_arguments(args: ast.arguments) -> str:
        """Format the argument list of a function into a string."""
        formatted_args: list[str] = []

        # Filter out 'self' parameter
        total_args = [arg for arg in args.args if arg.arg != "self"]
        defaults = args.defaults
        num_defaults = len(defaults)
        num_args = len(total_args)

        # Handle regular arguments with defaults
        for i, arg in enumerate(total_args):
            has_default = i >= num_args - num_defaults
            default_value = (
                defaults[i - (num_args - num_defaults)] if has_default else None
            )

            arg_str = arg.arg
            if arg.annotation:
                arg_str += f": {ast.unparse(arg.annotation)}"
            if default_value is not None:
                arg_str += f" = {ast.unparse(default_value)}"

            formatted_args.append(arg_str)

        # Handle *args
        if args.vararg:
            vararg_str = f"*{args.vararg.arg}"
            if args.vararg.annotation:
                vararg_str += f": {ast.unparse(args.vararg.annotation)}"
            formatted_args.append(vararg_str)

        return ", ".join(formatted_args)

    @staticmethod
    def extract_feature_key(decorator_node: ast.AST) -> FeatureKey | None:
        """Extract the feature key string from a decorator AST node."""
        if not isinstance(decorator_node, ast.Call) or not decorator_node.args:
            return None

        arg = decorator_node.args[0]
        if not isinstance(arg, ast.Constant):
            return None

        value = arg.value
        if isinstance(value, str) and value in FeatureKey.__args__:
            return value  # pyright: ignore[reportReturnType]

        return None

    @classmethod
    def should_include_function(cls, func: ast.FunctionDef) -> bool:
        """Determine whether a function should be included based on feature flags."""
        for decorator in func.decorator_list:
            feature_key = cls.extract_feature_key(decorator)
            if feature_key is not None and not has_feature(feature_key):
                return False
        return True

    @classmethod
    def parse_function(cls, func: ast.FunctionDef) -> FunctionStub:
        """Parse a function definition into a FunctionStub."""
        signature = (
            f"def {func.name}({cls.format_arguments(func.args)}) -> "
            f"{cls.format_return_annotation(func.returns)}:"
        )
        docstring = ast.get_docstring(func)

        return FunctionStub(name=func.name, signature=signature, docstring=docstring)


class ResourceReader:
    """Handles reading and parsing of Python source files."""

    @staticmethod
    def read_source(package: str, filename: str) -> str | None:
        """Read source code from a package resource."""
        try:
            return resources.read_text(package, filename)
        except (FileNotFoundError, OSError) as e:
            print(f"Error reading {filename} from {package}: {e}")
            return None

    @classmethod
    def get_all_exports(cls) -> list[str]:
        """Read and parse __init__.py from the `aegis_game` package to extract __all__."""
        source = cls.read_source("aegis_game", "__init__.py")
        if source is None:
            return []

        tree = ast.parse(source)
        for node in tree.body:
            if not isinstance(node, ast.Assign):
                continue

            for target in node.targets:
                if not (isinstance(target, ast.Name) and target.id == "__all__"):
                    continue

                if not isinstance(node.value, ast.List):
                    continue

                return [
                    elt.value
                    for elt in node.value.elts
                    if isinstance(elt, ast.Constant)
                    and isinstance(elt.value, str)
                    and elt.value != "main"
                ]

        print("Warning: __all__ not found in __init__.py")
        return []

    @classmethod
    def get_method_names(cls) -> list[str]:
        """Extract method names from the game.py methods function."""
        source = cls.read_source("_aegis_game", "game.py")
        if source is None:
            return []

        tree = ast.parse(source)

        for node in tree.body:
            if not isinstance(node, ast.ClassDef):
                continue

            for func in node.body:
                if not (isinstance(func, ast.FunctionDef) and func.name == "methods"):
                    continue

                return cls._extract_method_names_from_function(func)

        return []

    @staticmethod
    def _extract_method_names_from_function(func: ast.FunctionDef) -> list[str]:
        """Extract method names from a methods function that returns a dict."""
        methods: list[str] = []

        for node in ast.walk(func):
            if not (isinstance(node, ast.Return) and isinstance(node.value, ast.Dict)):
                continue

            methods.extend(
                key.value
                for key in node.value.keys
                if isinstance(key, ast.Constant)
                and isinstance(key.value, str)
                and not key.value[0].isupper()
            )

        return methods

    @classmethod
    def get_functions_from_module(
        cls, package: str, module: str, method_names: list[str]
    ) -> list[ast.FunctionDef]:
        """Get function definitions from a specific module."""
        source = cls.read_source(package, module)
        if source is None:
            return []

        tree = ast.parse(source)
        return [
            node
            for node in ast.walk(tree)
            if isinstance(node, ast.FunctionDef) and node.name in method_names
        ]


class StubGenerator:
    """Main class for generating stub files."""

    def __init__(self) -> None:
        self.parser: ASTFunctionParser = ASTFunctionParser()
        self.reader: ResourceReader = ResourceReader()

    def _build_header(self, stubs: list[FunctionStub]) -> str:
        """Generate the import/header section for the stub file."""
        needs_numpy = any("predict" in stub.name for stub in stubs)
        needs_messages = any("message" in stub.name for stub in stubs)

        imports: list[str] = []
        if needs_numpy:
            imports.extend(["import numpy as np", "from numpy.typing import NDArray"])

        relative_imports = self.reader.get_all_exports()
        if not needs_messages and "Message" in relative_imports:
            relative_imports.remove("Message")

        relative_imports.sort()

        import_lines: list[str] = []
        import_lines.extend(imports)
        import_lines.append("from . import (")
        import_lines.extend(f"    {name}," for name in relative_imports)
        import_lines.append(")\n\n")

        import_block = "\n".join(import_lines)

        return f'''"""
Autogenerated from API functions.

Do not modify manually.
"""
# ruff: noqa: F401
# pyright: reportReturnType=false
# pyright: reportUnusedImport=false
# pyright: reportUnusedParameter=false

{import_block}
'''

    def _collect_functions(self) -> list[ast.FunctionDef]:
        """Collect all relevant function definitions from modules."""
        method_names = self.reader.get_method_names()

        # Get functions from agent_controller.py
        ac_functions = self.reader.get_functions_from_module(
            "_aegis_game", "agent_controller.py", method_names
        )
        ac_names = {func.name for func in ac_functions}

        # Get functions from game.py (excluding those already in agent_controller)
        game_functions = self.reader.get_functions_from_module(
            "_aegis_game", "game.py", method_names
        )
        game_functions = [func for func in game_functions if func.name not in ac_names]

        return ac_functions + game_functions

    def _write_stub_file(self, content: str) -> None:
        """Write the generated content to the stub file."""
        output_path = Path(__file__).parent.parent.parent / "aegis_game" / "stub.py"
        try:
            _ = output_path.write_text(content, encoding="utf-8", newline="\n")
            print("Successfully generated stub.")
        except (OSError, PermissionError) as e:
            print(f"Error writing stub.py: {e}")

    def generate(self) -> None:
        """Generate the complete stub file."""
        all_functions = self._collect_functions()

        # Filter functions based on feature flags
        included_functions = [
            func for func in all_functions if self.parser.should_include_function(func)
        ]

        # Convert to function stubs
        stubs = [self.parser.parse_function(func) for func in included_functions]

        # Generate content
        header = self._build_header(stubs)
        stub_content = "\n\n".join(
            stub.format(is_last=(i == len(stubs) - 1)) for i, stub in enumerate(stubs)
        )

        content = header + stub_content + "\n"

        # Write to file
        self._write_stub_file(content)
        print(f"Generated stub with {len(stubs)} functions")


def main() -> None:
    """Entry point for the stub generator."""
    generator = StubGenerator()
    generator.generate()


if __name__ == "__main__":
    main()



from . import location_pb2 as _location_pb2
from . import team_pb2 as _team_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Spawn(_message.Message):
    __slots__ = ("agentId", "loc", "team")
    AGENTID_FIELD_NUMBER: _ClassVar[int]
    LOC_FIELD_NUMBER: _ClassVar[int]
    TEAM_FIELD_NUMBER: _ClassVar[int]
    agentId: int
    loc: _location_pb2.Location
    team: _team_pb2.Team
    def __init__(self, agentId: _Optional[int] = ..., loc: _Optional[_Union[_location_pb2.Location, _Mapping]] = ..., team: _Optional[_Union[_team_pb2.Team, str]] = ...) -> None: ...

class InitSpawn(_message.Message):
    __slots__ = ("loc", "amount")
    LOC_FIELD_NUMBER: _ClassVar[int]
    AMOUNT_FIELD_NUMBER: _ClassVar[int]
    loc: _location_pb2.Location
    amount: int
    def __init__(self, loc: _Optional[_Union[_location_pb2.Location, _Mapping]] = ..., amount: _Optional[int] = ...) -> None: ...



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SurvivorState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ALIVE: _ClassVar[SurvivorState]
    DEAD: _ClassVar[SurvivorState]
ALIVE: SurvivorState
DEAD: SurvivorState

class Survivor(_message.Message):
    __slots__ = ("id", "health", "state")
    ID_FIELD_NUMBER: _ClassVar[int]
    HEALTH_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    id: int
    health: int
    state: SurvivorState
    def __init__(self, id: _Optional[int] = ..., health: _Optional[int] = ..., state: _Optional[_Union[SurvivorState, str]] = ...) -> None: ...

class Rubble(_message.Message):
    __slots__ = ("id", "energy_required", "agents_required")
    ID_FIELD_NUMBER: _ClassVar[int]
    ENERGY_REQUIRED_FIELD_NUMBER: _ClassVar[int]
    AGENTS_REQUIRED_FIELD_NUMBER: _ClassVar[int]
    id: int
    energy_required: int
    agents_required: int
    def __init__(self, id: _Optional[int] = ..., energy_required: _Optional[int] = ..., agents_required: _Optional[int] = ...) -> None: ...

class WorldObject(_message.Message):
    __slots__ = ("survivor", "rubble")
    SURVIVOR_FIELD_NUMBER: _ClassVar[int]
    RUBBLE_FIELD_NUMBER: _ClassVar[int]
    survivor: Survivor
    rubble: Rubble
    def __init__(self, survivor: _Optional[_Union[Survivor, _Mapping]] = ..., rubble: _Optional[_Union[Rubble, _Mapping]] = ...) -> None: ...



"""
Script to populate the template directories for aegis init command.

This script copies the necessary files into the templates directory structure:
- templates/path/ - for pathfinding assignment
- templates/mas/ - for multi-agent assignment
- templates/comp/ - for competition

Run this script manually or integrate it into your build process.
"""

import shutil
import sys
from pathlib import Path

TEMPLATE_STRUCTURE: dict[str, dict[str, str | None]] = {
    "path": {
        "agents": "agents/agent_path",
        "worlds": "worlds/path-worlds",
        "config": "config/presets/pathfinding-assignment.yaml",
        "predictions": None,
    },
    "mas": {
        "agents": "agents/agent_mas",
        "worlds": "worlds/mas-worlds",
        "config": "config/presets/multi-agent-assignment.yaml",
        "predictions": "prediction_data",
    },
    "comp": {
        "agents": "agents/agent_comp",
        "worlds": "worlds/comp-worlds",
        "config": "config/presets/competiton-versus.yaml",
        "predictions": "prediction_data",
    },
}


def get_project_root() -> Path:
    """Get the project root directory."""
    current = Path.cwd()
    while current != current.parent:
        if (current / "pyproject.toml").exists():
            return current
        current = current.parent
    msg = "Could not find project root (pyproject.toml)"
    raise FileNotFoundError(msg)


def copy_directory_excluding_cache(src: Path, dest: Path) -> None:
    """Copy directory excluding __pycache__ and other cache files."""
    if dest.exists():
        shutil.rmtree(dest)

    def ignore_patterns(_: str, names: list[str]) -> list[str]:
        """Ignore __pycache__, .pyc, and other cache files."""
        return [
            name
            for name in names
            if name in {"__pycache__", ".pyc", ".pyo", ".pyd"}
            or name.endswith((".pyc", ".pyo", ".pyd"))
        ]

    _ = shutil.copytree(src, dest, ignore=ignore_patterns)


def _copy_agents(
    agents_path: str | None, project_root: Path, template_dir: Path
) -> None:
    """Copy agents directory to template."""
    if not agents_path:
        print("No agents specified")
        return

    agents_src = project_root / agents_path
    if not agents_src.exists():
        print(f"Warning: {agents_path} not found")
        return

    agents_dest = template_dir / "agents"
    agents_dest.mkdir(parents=True, exist_ok=True)

    # Extract the agent folder name (e.g., "agent_path" from "agents/agent_path")
    agent_folder_name = agents_path.split("/")[-1]
    agent_dest = agents_dest / agent_folder_name
    copy_directory_excluding_cache(agents_src, agent_dest)
    # print(f"Copied agents: {agents_path}")


def _copy_worlds(
    worlds_path: str | None, project_root: Path, template_dir: Path
) -> None:
    """Copy worlds directory to template."""
    if not worlds_path:
        print("No worlds specified")
        return

    worlds_src = project_root / worlds_path
    if not worlds_src.exists():
        print(f"Warning: {worlds_path} not found")
        return

    worlds_dest = template_dir / "worlds"
    if worlds_dest.exists():
        shutil.rmtree(worlds_dest)
    worlds_dest.mkdir(parents=True, exist_ok=True)

    # Copy entire directory
    copy_directory_excluding_cache(worlds_src, worlds_dest)
    # print(f"Copied worlds: {worlds_path}")


def _copy_predictions(
    predictions_path: str | None, project_root: Path, template_dir: Path
) -> None:
    """Copy prediction data to template."""
    if not predictions_path:
        print("No predictions specified")
        return

    predictions_src = project_root / predictions_path
    if not predictions_src.exists():
        print(f"Warning: {predictions_path} not found")
        return

    predictions_dest = template_dir / "prediction_data"
    copy_directory_excluding_cache(predictions_src, predictions_dest)
    # print(f"Copied prediction data: {predictions_path}")


def _copy_config(
    config_path: str | None, project_root: Path, template_dir: Path
) -> None:
    """Copy and rename config file to template."""
    if not config_path:
        print("No config specified")
        return

    config_src = project_root / config_path
    if not config_src.exists():
        print(f"Warning: {config_path} not found")
        return

    config_dest = template_dir / "config"
    config_dest.mkdir(parents=True, exist_ok=True)
    _ = shutil.copy2(config_src, config_dest / "config.yaml")
    # print(f"Copied config: {config_path} -> config.yaml")


def populate_template(
    template_name: str, template_config: dict[str, str | None], project_root: Path
) -> None:
    """Populate a single template directory."""
    template_dir = project_root / "src" / "_aegis_game" / "templates" / template_name

    if template_dir.exists():
        shutil.rmtree(template_dir)
    template_dir.mkdir(parents=True, exist_ok=True)

    # Copy all template components
    _copy_agents(template_config["agents"], project_root, template_dir)
    _copy_worlds(template_config["worlds"], project_root, template_dir)
    _copy_predictions(template_config["predictions"], project_root, template_dir)
    _copy_config(template_config["config"], project_root, template_dir)

    print(f"Template {template_name} populated successfully")


def main() -> None:
    """Populate templates for aegis init command."""
    try:
        project_root = get_project_root()

        # Clear the entire templates directory before populating
        templates_dir = project_root / "src" / "_aegis_game" / "templates"
        if templates_dir.exists():
            print("Clearing existing templates directory...")
            shutil.rmtree(templates_dir)
        templates_dir.mkdir(parents=True, exist_ok=True)

        for template_name, template_config in TEMPLATE_STRUCTURE.items():
            populate_template(template_name, template_config, project_root)

        print("\nAll templates populated successfully!")
        print("\nTemplate structure created:")
        for template_name in TEMPLATE_STRUCTURE:
            template_dir = project_root / "src" / "_aegis" / "templates" / template_name
            print(f"  - {template_name}: {template_dir}")

    except Exception as e:  # noqa: BLE001
        print(f"Error populating templates: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import argparse
from dataclasses import dataclass

from .aegis_config import get_feature_value
from .constants import Constants


@dataclass
class TypedNamespace:
    command: str
    amount: int
    world: list[str]
    rounds: int
    agent: str | None
    agent2: str | None
    client: bool
    debug: bool
    log: bool
    init_type: str


@dataclass
class LaunchArgs:
    amount: int
    world: list[str]
    rounds: int
    agent: str | None
    agent2: str | None
    client: bool
    debug: bool
    log: bool


@dataclass
class InitArgs:
    init_type: str


@dataclass
class ForgeArgs:
    # If we ever need args
    pass


@dataclass
class Args:
    command: str
    launch_args: LaunchArgs | None = None
    forge_args: ForgeArgs | None = None
    init_args: InitArgs | None = None


def parse_args() -> Args:
    parser = argparse.ArgumentParser(description="AEGIS Simulation")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # Do not require a config file for commands other than launch.
    # If the config is missing, we will fall back to a safe default of 1.
    try:
        default_agent_amount = get_feature_value("DEFAULT_AGENT_AMOUNT")
    except FileNotFoundError:
        default_agent_amount = None

    run_parser = subparsers.add_parser("launch", help="Run a game")
    _ = run_parser.add_argument(
        "--world",
        type=str,
        nargs="+",
        required=True,
        help="One or more world names (without .world extension), separated by spaces.",
    )
    _ = run_parser.add_argument(
        "--amount",
        type=int,
        default=default_agent_amount if default_agent_amount is not None else 1,
        help="Number of agents to run (default = 1)",
    )
    _ = run_parser.add_argument(
        "--rounds",
        type=int,
        default=Constants.DEFAULT_MAX_ROUNDS,
        help=f"Number of simulation rounds (default = {Constants.DEFAULT_MAX_ROUNDS})",
    )
    _ = run_parser.add_argument(
        "--agent",
        type=str,
        required=False,
        help=(
            "Name of the agent folder under 'agents/' with a main.py file "
            "for team Goobs (e.g., 'agent_path', 'agent_mas')"
        ),
    )
    _ = run_parser.add_argument(
        "--agent2",
        type=str,
        required=False,
        help=(
            "Name of the agent folder under 'agents/' with a main.py file "
            "for team Voidseers (e.g., 'agent_path', 'agent_mas')"
        ),
    )
    _ = run_parser.add_argument(
        "--client",
        action="store_true",
        help="Used by the client, tells the server to wait for the client to connect",
    )
    _ = run_parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable agent debug logging output",
    )
    _ = run_parser.add_argument(
        "--log",
        action="store_true",
        help="Enable AEGIS console output logging to a file",
    )

    _ = subparsers.add_parser("forge", help="Make stub.py file after config changes")

    init_parser = subparsers.add_parser(
        "init",
        help=("Create required folders/files to run a simulation"),
    )
    _ = init_parser.add_argument(
        "--type",
        dest="init_type",
        choices=["path", "mas", "comp"],
        default="path",
        help="Initialization type: 'path' (default), 'mas', or 'comp'",
    )

    args = parser.parse_args(namespace=TypedNamespace)

    if args.command == "launch":
        return Args(
            command="run",
            launch_args=LaunchArgs(
                amount=args.amount,
                world=args.world,
                rounds=args.rounds,
                agent=args.agent,
                agent2=args.agent2,
                client=args.client,
                debug=args.debug,
                log=args.log,
            ),
        )
    if args.command == "forge":
        return Args(command="forge", forge_args=ForgeArgs())
    if args.command == "init":
        return Args(command="init", init_args=InitArgs(init_type=args.init_type))

    error = f"Unknown command: {args.command}"
    raise ValueError(error)



import sys
import traceback

from .args_parser import parse_args
from .cli_scripts.client_installer import main as install_client
from .cli_scripts.init_scaffold import init_scaffold
from .play import run


def main() -> None:
    args = parse_args()

    if args.command == "run":
        try:
            if args.launch_args is None:
                sys.exit(1)
            run(args.launch_args)
        except Exception as e:  # noqa: BLE001
            print(f"ERROR: {e}", file=sys.stderr)
            traceback.print_exc()
            sys.exit(1)

    elif args.command == "forge":
        from .cli_scripts.build_public_api import main as build_api  # noqa: PLC0415

        try:
            build_api()
        except Exception:  # noqa: BLE001
            traceback.print_exc()
            sys.exit(1)
    elif args.command == "init":
        try:
            if args.init_args is None:
                sys.exit(1)
            init_scaffold(args.init_args.init_type)
            install_client()
        except Exception:  # noqa: BLE001
            traceback.print_exc()
            sys.exit(1)


if __name__ == "__main__":
    main()



from collections.abc import Callable
from typing import TypeVar

from _aegis_game.types.others import FeatureKey

F = TypeVar("F", bound=Callable[..., object])


def requires(_feature_key: FeatureKey) -> Callable[[F], F]:
    def decorator(func: F) -> F:
        return func

    return decorator



from typing import TypedDict

import numpy as np
from numpy.typing import NDArray

from _aegis_game.team import Team


# Type for pending predictions
class PendingPrediction(TypedDict):
    image_to_predict: NDArray[np.uint8]
    correct_label: np.int32


# Type for completed predictions
class CompletedPrediction(TypedDict):
    team: Team
    surv_id: int
    is_correct: bool



import logging
import sys
from datetime import UTC, datetime
from pathlib import Path
from typing import override


class MaxLevelFilter(logging.Filter):
    def __init__(self, max_level: int) -> None:
        super().__init__()
        self.max_level: int = max_level

    @override
    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno <= self.max_level


class LevelBasedFormatter(logging.Formatter):
    @override
    def format(self, record: logging.LogRecord) -> str:
        if record.name == "aegis":
            if record.levelno >= logging.WARNING:
                self._style._fmt = "[aegis:%(levelname)s] %(message)s"  # noqa: SLF001
            else:
                self._style._fmt = "[aegis] %(message)s"  # noqa: SLF001
        elif record.name == "aegis.agent":
            if record.levelno >= logging.WARNING:
                return f"[Agent] {record.getMessage()}"
            return record.getMessage()
        return super().format(record)


def setup_console_logging() -> None:
    """Set up basic console logging without file output."""
    formatter: logging.Formatter = LevelBasedFormatter()

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.addFilter(MaxLevelFilter(logging.INFO))
    stdout_handler.setFormatter(formatter)

    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    stderr_handler.setLevel(logging.WARNING)
    stderr_handler.setFormatter(formatter)

    handlers: list[logging.Handler] = [stdout_handler, stderr_handler]

    logging.basicConfig(
        level=logging.INFO,
        handlers=handlers,
        force=True,
    )

    # Configure agent logger to use stdout for INFO/DEBUG and stderr for WARNING/ERROR
    agent_logger = logging.getLogger("aegis.agent")
    agent_logger.setLevel(logging.DEBUG)
    agent_logger.propagate = False  # Prevent double logging

    # Handler for normal messages (stdout)
    agent_stdout_handler = logging.StreamHandler(stream=sys.stdout)
    agent_stdout_handler.setLevel(logging.DEBUG)
    agent_stdout_handler.addFilter(lambda record: record.levelno < logging.WARNING)
    agent_stdout_handler.setFormatter(formatter)
    agent_logger.addHandler(agent_stdout_handler)

    # Handler for error messages (stderr)
    agent_stderr_handler = logging.StreamHandler(stream=sys.stderr)
    agent_stderr_handler.setLevel(logging.WARNING)
    agent_stderr_handler.setFormatter(formatter)
    agent_logger.addHandler(agent_stderr_handler)


def setup_console_and_file_logging() -> None:
    """Set up AEGIS logging with both console and file output."""
    formatter: logging.Formatter = LevelBasedFormatter()

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.addFilter(MaxLevelFilter(logging.INFO))
    stdout_handler.setFormatter(formatter)

    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    stderr_handler.setLevel(logging.WARNING)
    stderr_handler.setFormatter(formatter)

    logs_dir: Path = Path.cwd() / "logs"
    logs_dir.mkdir(exist_ok=True)

    timestamp: str = datetime.now(UTC).strftime("%Y%m%d_%H%M%S")
    log_filename: str = f"aegis_simulation_{timestamp}.log"
    log_file_path: Path = logs_dir / log_filename

    file_handler: logging.FileHandler = logging.FileHandler(
        log_file_path, mode="w", encoding="utf-8"
    )
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(
        logging.Formatter("[%(asctime)s][%(levelname)s][%(name)s] - %(message)s")
    )

    handlers: list[logging.Handler] = [stdout_handler, stderr_handler, file_handler]

    logging.basicConfig(
        level=logging.INFO,
        handlers=handlers,
        force=True,
    )

    # Configure agent logger to use stdout for INFO/DEBUG and stderr for WARNING/ERROR
    agent_logger = logging.getLogger("aegis.agent")
    agent_logger.setLevel(logging.DEBUG)
    agent_logger.propagate = False  # Prevent double logging

    # Handler for normal messages (stdout)
    agent_stdout_handler = logging.StreamHandler(stream=sys.stdout)
    agent_stdout_handler.setLevel(logging.DEBUG)
    agent_stdout_handler.addFilter(lambda record: record.levelno < logging.WARNING)
    agent_stdout_handler.setFormatter(formatter)
    agent_logger.addHandler(agent_stdout_handler)

    # Handler for error messages (stderr)
    agent_stderr_handler = logging.StreamHandler(stream=sys.stderr)
    agent_stderr_handler.setLevel(logging.WARNING)
    agent_stderr_handler.setFormatter(formatter)
    agent_logger.addHandler(agent_stderr_handler)


LOGGER: logging.Logger = logging.getLogger("aegis")
AGENT_LOGGER: logging.Logger = logging.getLogger("aegis.agent")



"""Tests for the Direction enum."""

from __future__ import annotations

import pytest

from src._aegis_game.common import Direction


class TestBasics:
    """Test basic Direction enum functionality."""

    def test_direction_values(self) -> None:
        """Test that each direction has the correct (dx, dy) values."""
        assert Direction.NORTH.value == (0, 1)
        assert Direction.NORTHEAST.value == (1, 1)
        assert Direction.EAST.value == (1, 0)
        assert Direction.SOUTHEAST.value == (1, -1)
        assert Direction.SOUTH.value == (0, -1)
        assert Direction.SOUTHWEST.value == (-1, -1)
        assert Direction.WEST.value == (-1, 0)
        assert Direction.NORTHWEST.value == (-1, 1)
        assert Direction.CENTER.value == (0, 0)

    def test_dx_property(self) -> None:
        """Test the dx property returns correct x-axis change."""
        assert Direction.NORTH.dx == 0
        assert Direction.NORTHEAST.dx == 1
        assert Direction.EAST.dx == 1
        assert Direction.SOUTHEAST.dx == 1
        assert Direction.SOUTH.dx == 0
        assert Direction.SOUTHWEST.dx == -1
        assert Direction.WEST.dx == -1
        assert Direction.NORTHWEST.dx == -1
        assert Direction.CENTER.dx == 0

    def test_dy_property(self) -> None:
        """Test the dy property returns correct y-axis change."""
        assert Direction.NORTH.dy == 1
        assert Direction.NORTHEAST.dy == 1
        assert Direction.EAST.dy == 0
        assert Direction.SOUTHEAST.dy == -1
        assert Direction.SOUTH.dy == -1
        assert Direction.SOUTHWEST.dy == -1
        assert Direction.WEST.dy == 0
        assert Direction.NORTHWEST.dy == 1
        assert Direction.CENTER.dy == 0


class TestRotation:
    """Test direction rotation methods."""

    @pytest.mark.parametrize(
        ("direction", "expected"),
        [
            (Direction.NORTH, Direction.NORTHWEST),
            (Direction.NORTHEAST, Direction.NORTH),
            (Direction.EAST, Direction.NORTHEAST),
            (Direction.SOUTHEAST, Direction.EAST),
            (Direction.SOUTH, Direction.SOUTHEAST),
            (Direction.SOUTHWEST, Direction.SOUTH),
            (Direction.WEST, Direction.SOUTHWEST),
            (Direction.NORTHWEST, Direction.WEST),
            (Direction.CENTER, Direction.CENTER),
        ],
    )
    def test_rotate_left(self, direction: Direction, expected: Direction) -> None:
        """Test rotating left (counter-clockwise) 45 degrees."""
        assert direction.rotate_left() == expected

    @pytest.mark.parametrize(
        ("direction", "expected"),
        [
            (Direction.NORTH, Direction.NORTHEAST),
            (Direction.NORTHEAST, Direction.EAST),
            (Direction.EAST, Direction.SOUTHEAST),
            (Direction.SOUTHEAST, Direction.SOUTH),
            (Direction.SOUTH, Direction.SOUTHWEST),
            (Direction.SOUTHWEST, Direction.WEST),
            (Direction.WEST, Direction.NORTHWEST),
            (Direction.NORTHWEST, Direction.NORTH),
            (Direction.CENTER, Direction.CENTER),
        ],
    )
    def test_rotate_right(self, direction: Direction, expected: Direction) -> None:
        """Test rotating right (clockwise) 45 degrees."""
        assert direction.rotate_right() == expected

    def test_multiple_rotations(self) -> None:
        """Test that 8 rotations in one direction return to original."""
        direction = Direction.NORTH

        # 8 left rotations should return to start
        current = direction
        for _ in range(8):
            current = current.rotate_left()
        assert current == direction

        # 8 right rotations should return to start
        current = direction
        for _ in range(8):
            current = current.rotate_right()
        assert current == direction

    def test_left_right_inverse(self) -> None:
        """Test that rotating left then right returns to original."""
        for direction in [d for d in Direction if d != Direction.CENTER]:
            rotated_left = direction.rotate_left()
            back_to_original = rotated_left.rotate_right()
            assert back_to_original == direction


class TestOpposite:
    """Test the get_opposite method."""

    @pytest.mark.parametrize(
        ("direction", "expected"),
        [
            (Direction.NORTH, Direction.SOUTH),
            (Direction.NORTHEAST, Direction.SOUTHWEST),
            (Direction.EAST, Direction.WEST),
            (Direction.SOUTHEAST, Direction.NORTHWEST),
            (Direction.SOUTH, Direction.NORTH),
            (Direction.SOUTHWEST, Direction.NORTHEAST),
            (Direction.WEST, Direction.EAST),
            (Direction.NORTHWEST, Direction.SOUTHEAST),
            (Direction.CENTER, Direction.CENTER),
        ],
    )
    def test_get_opposite(self, direction: Direction, expected: Direction) -> None:
        """Test that get_opposite returns the correct opposite direction."""
        assert direction.get_opposite() == expected

    def test_opposite_symmetry(self) -> None:
        """Test that the opposite of an opposite is the original."""
        for direction in Direction:
            opposite = direction.get_opposite()
            opposite_of_opposite = opposite.get_opposite()
            assert opposite_of_opposite == direction


class TestEdgeCases:
    """Test edge cases and error conditions."""

    def test_all_directions_exist(self) -> None:
        """Test that all expected directions exist in the enum."""
        expected_directions = {
            "NORTH",
            "NORTHEAST",
            "EAST",
            "SOUTHEAST",
            "SOUTH",
            "SOUTHWEST",
            "WEST",
            "NORTHWEST",
            "CENTER",
        }
        actual_directions = {d.name for d in Direction}
        assert actual_directions == expected_directions

    def test_direction_count(self) -> None:
        """Test that there are exactly 9 directions."""
        dir_expected_length: int = 9
        assert len(Direction) == dir_expected_length

    def test_enum_iteration(self) -> None:
        """Test that we can iterate over all directions."""
        directions = list(Direction)
        dir_expected_length: int = 9
        assert len(directions) == dir_expected_length
        assert Direction.NORTH in directions
        assert Direction.CENTER in directions



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Team(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    GOOBS: _ClassVar[Team]
    VOIDSEERS: _ClassVar[Team]
GOOBS: Team
VOIDSEERS: Team

class TeamInfo(_message.Message):
    __slots__ = ("saved_alive", "saved_dead", "saved", "predicted_right", "predicted_wrong", "predicted", "score", "units", "team")
    SAVED_ALIVE_FIELD_NUMBER: _ClassVar[int]
    SAVED_DEAD_FIELD_NUMBER: _ClassVar[int]
    SAVED_FIELD_NUMBER: _ClassVar[int]
    PREDICTED_RIGHT_FIELD_NUMBER: _ClassVar[int]
    PREDICTED_WRONG_FIELD_NUMBER: _ClassVar[int]
    PREDICTED_FIELD_NUMBER: _ClassVar[int]
    SCORE_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    TEAM_FIELD_NUMBER: _ClassVar[int]
    saved_alive: int
    saved_dead: int
    saved: int
    predicted_right: int
    predicted_wrong: int
    predicted: int
    score: int
    units: int
    team: Team
    def __init__(self, saved_alive: _Optional[int] = ..., saved_dead: _Optional[int] = ..., saved: _Optional[int] = ..., predicted_right: _Optional[int] = ..., predicted_wrong: _Optional[int] = ..., predicted: _Optional[int] = ..., score: _Optional[int] = ..., units: _Optional[int] = ..., team: _Optional[_Union[Team, str]] = ...) -> None: ...



import random


class IDGenerator:
    def __init__(self, start: int = 10001, count: int = 4096) -> None:
        self.available_ids: list[int] = list(range(start, start + count))
        random.shuffle(self.available_ids)

    def next_id(self) -> int:
        if not self.available_ids:
            error = "No IDs left to allocate"
            raise RuntimeError(error)
        return self.available_ids.pop()



from pathlib import Path

from google.protobuf.message import DecodeError

from .aegis_config import has_feature
from .args_parser import LaunchArgs
from .game import Game
from .game_pb import GamePb
from .logger import LOGGER, setup_console_and_file_logging, setup_console_logging
from .sandbox.sandbox import Sandbox
from .server_websocket import WebSocketServer
from .team import Team
from .world_pb import load_world


def log_game_end(game: Game, args: LaunchArgs, i: int) -> None:
    LOGGER.info("========== AEGIS END ==========")
    LOGGER.info(f"Finished on round {game.round}")
    LOGGER.info(f"Reason: {getattr(game.reason, 'value', 'Unknown')}")
    LOGGER.info("")
    LOGGER.info(f"{'Team':<12} {'Score':>8} {'Saved':>8}")
    if has_feature("ALLOW_AGENT_PREDICTIONS"):
        LOGGER.info(f"{'Team':<12} {'Score':>8} {'Saved':>8} {'Predictions':>14}")
        LOGGER.info("-" * 46)
    else:
        LOGGER.info("-" * 31)

    for team in Team:
        if team == Team.VOIDSEERS and args.agent2 is None:
            continue

        score = game.team_info.get_score(team)
        saved = game.team_info.get_saved(team)
        predictions = game.team_info.get_predicted_right(team)

        if has_feature("ALLOW_AGENT_PREDICTIONS"):
            LOGGER.info(f"{team.name:<12} {score:>8} {saved:>8} {predictions:>14}")
        else:
            LOGGER.info(f"{team.name:<12} {score:>8} {saved:>8}")

    if has_feature("ALLOW_AGENT_PREDICTIONS"):
        LOGGER.info("=" * 46)
    else:
        LOGGER.info("=" * 31)

    if i < len(args.world) - 1:
        print("\n" + "=" * 80 + "\n\n")


def make_game_start_string(args: LaunchArgs, world: str) -> str:
    if args.agent and not args.agent2:
        return f"GOOBS on {world}"

    if args.agent2 and not args.agent:
        return f"VOIDSEERS on {world}"

    # TODO @dante: This will have to show actual team names
    return f"GOOBS vs VOIDSEERS on {world}"


def run(args: LaunchArgs) -> None:
    if args.agent is None and args.agent2 is None:
        error = "At least one agent must be provided"
        raise ValueError(error)

    setup_console_and_file_logging() if args.log else setup_console_logging()

    sandbox_goobs = (
        Sandbox.from_directory(Path.cwd() / "agents" / args.agent)
        if args.agent is not None
        else None
    )
    sandbox_seers = (
        Sandbox.from_directory(Path.cwd() / "agents" / args.agent2)
        if args.agent2 is not None
        else None
    )
    ws_server = WebSocketServer(wait_for_client=args.client)
    game_pb = GamePb()

    ws_server.start()
    game_pb.make_games_header(ws_server)

    for i, arg_world in enumerate(args.world):
        world_name = f"{arg_world}"
        world_path = Path.cwd() / "worlds" / f"{world_name}.world"

        try:
            world = load_world(world_path)
        except (FileNotFoundError, DecodeError) as e:
            error = f"Unable to load world {world_path}!"
            raise ValueError(error) from e

        world.rounds = args.rounds

        try:
            game = Game([sandbox_goobs, sandbox_seers], args, world, game_pb)
        except ValueError as e:
            enhanced_msg = f"Error in world '{world_name}': {e}"
            raise ValueError(enhanced_msg) from e

        LOGGER.info("========== AEGIS START ==========")
        LOGGER.info(make_game_start_string(args, world_name))

        game_pb.make_game_header(world)
        while game.running:
            try:
                game.run_round()
            except Exception:  # noqa: BLE001
                LOGGER.exception("This shouldn't have happened. Internal error.")
                game.running = False

        game_pb.make_game_footer()
        log_game_end(game, args, i)
    game_pb.make_games_footer()
    ws_server.finish()



from . import cell_pb2 as _cell_pb2
from . import spawn_pb2 as _spawn_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class World(_message.Message):
    __slots__ = ("width", "height", "seed", "start_energy", "cells", "total_survivors", "init_spawns")
    WIDTH_FIELD_NUMBER: _ClassVar[int]
    HEIGHT_FIELD_NUMBER: _ClassVar[int]
    SEED_FIELD_NUMBER: _ClassVar[int]
    START_ENERGY_FIELD_NUMBER: _ClassVar[int]
    CELLS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SURVIVORS_FIELD_NUMBER: _ClassVar[int]
    INIT_SPAWNS_FIELD_NUMBER: _ClassVar[int]
    width: int
    height: int
    seed: int
    start_energy: int
    cells: _containers.RepeatedCompositeFieldContainer[_cell_pb2.Cell]
    total_survivors: int
    init_spawns: _containers.RepeatedCompositeFieldContainer[_spawn_pb2.InitSpawn]
    def __init__(self, width: _Optional[int] = ..., height: _Optional[int] = ..., seed: _Optional[int] = ..., start_energy: _Optional[int] = ..., cells: _Optional[_Iterable[_Union[_cell_pb2.Cell, _Mapping]]] = ..., total_survivors: _Optional[int] = ..., init_spawns: _Optional[_Iterable[_Union[_spawn_pb2.InitSpawn, _Mapping]]] = ...) -> None: ...



from pathlib import Path
from typing import TYPE_CHECKING, cast

import numpy as np

from _aegis_game.args_parser import LaunchArgs

if TYPE_CHECKING:
    from numpy.typing import NDArray


class PredictionDataLoader:
    """Handles loading prediction data from external directories."""

    def __init__(self, args: LaunchArgs) -> None:
        """
        Initialize the data loader.

        Args:
            args: The arguments object

        """
        self.args: LaunchArgs = args
        self.x_test: NDArray[np.uint8] = np.array([])
        self.y_test: NDArray[np.int32] = np.array([])
        self.unique_labels: NDArray[np.int32] = np.array([])
        self.load_testing_data()

    def load_testing_data(self) -> None:
        """Load testing data from the testing directory."""
        data_dir = Path.cwd() / "prediction_data" / "testing"

        x_path = data_dir / "x_test_symbols.npy"
        y_path = data_dir / "y_test_symbols.npy"

        if not x_path.exists() or not y_path.exists():
            msg = (
                f"Prediction data not found in {data_dir}. "
                f"Expected files: x_test_symbols.npy, y_test_symbols.npy"
            )
            raise FileNotFoundError(msg)

        self.x_test = cast("NDArray[np.uint8]", np.load(x_path))
        self.y_test = cast("NDArray[np.int32]", np.load(y_path))
        self.unique_labels = np.unique(self.y_test)

    @property
    def num_testing_images(self) -> int:
        """Get the number of testing images available."""
        return len(self.x_test)



from .agent import Agent
from .common import Location
from .schemas.event_pb2 import Event
from .schemas.game_pb2 import (
    DroneScan,
    DroneScanUpdate,
    GameFooter,
    GameHeader,
    GamesFooter,
    GamesHeader,
    Round,
    SurvivorHealthUpdate,
)
from .schemas.location_pb2 import Location as PbLocation
from .schemas.spawn_pb2 import Spawn
from .schemas.team_pb2 import Team as PbTeam
from .schemas.team_pb2 import TeamInfo as PbTeamInfo
from .schemas.turn_pb2 import Turn
from .schemas.world_object_pb2 import SurvivorState
from .server_websocket import WebSocketServer
from .team import Team
from .team_info import TeamInfo
from .world import World
from .world_pb import serialize_world


class GamePb:
    def __init__(self) -> None:
        self.round: int = 0
        self.team_info: list[PbTeamInfo] = []
        self.turns: list[Turn] = []
        self.spawns: list[Spawn] = []
        self.removed_layers: list[PbLocation] = []
        self.dead_ids: list[int] = []
        self.drone_scans: list[DroneScan] = []
        self.survivor_health_updates: list[SurvivorHealthUpdate] = []
        self.ws_server: WebSocketServer | None = None

    def make_games_header(self, ws_server: WebSocketServer) -> None:
        self.ws_server = ws_server
        games_header = GamesHeader()

        event = Event()
        event.games_header.CopyFrom(games_header)

        binary_string = event.SerializeToString()
        self.ws_server.add_event(binary_string)

    def make_game_header(self, world: World) -> None:
        if self.ws_server is None:
            error = "Server should have started."
            raise ValueError(error)

        game_header = GameHeader()
        pb_world = serialize_world(world)
        game_header.world.CopyFrom(pb_world)
        game_header.rounds = world.rounds

        if self.spawns:
            game_header.spawns.extend(self.spawns)

        event = Event()
        event.game_header.CopyFrom(game_header)

        binary_string = event.SerializeToString()
        self.ws_server.add_event(binary_string)
        # clear so it doesn't keep ids for agent turn spawns
        self.spawns.clear()

    def start_round(self, game_round: int) -> None:
        self.round = game_round

    def send_drone_scan_update(
        self, drone_scans: dict[Location, dict[Team, int]]
    ) -> None:
        """Send drone scan data to the client."""
        if self.ws_server is None:
            error = "Server should have started."
            raise ValueError(error)

        pb_drone_update = DroneScanUpdate()

        for loc, teams in drone_scans.items():
            for team, duration in teams.items():
                drone_scan = DroneScan()
                drone_scan.location.x = loc.x
                drone_scan.location.y = loc.y
                drone_scan.team = (
                    PbTeam.GOOBS if team == Team.GOOBS else PbTeam.VOIDSEERS
                )
                drone_scan.duration = duration
                pb_drone_update.drone_scans.append(drone_scan)

        event = Event()
        event.drone_scan_update.CopyFrom(pb_drone_update)

        binary_string = event.SerializeToString()
        self.ws_server.add_event(binary_string)

    def end_round(self) -> None:
        if self.ws_server is None:
            error = "Server should have started."
            raise ValueError(error)
        pb_round = Round()
        pb_round.round = self.round
        pb_round.turns.extend(self.turns)
        pb_round.team_info.extend(self.team_info)
        pb_round.layers_removed.extend(self.removed_layers)
        pb_round.dead_ids.extend(self.dead_ids)
        pb_round.drone_scans.extend(self.drone_scans)
        pb_round.survivor_health_updates.extend(self.survivor_health_updates)

        event = Event()
        event.round.CopyFrom(pb_round)

        binary_string = event.SerializeToString()
        self.ws_server.add_event(binary_string)
        self.clear_round()

    def end_turn(self, agent: Agent) -> None:
        pb_turn = Turn()
        pb_turn.agentId = agent.id
        pb_turn.energy_level = agent.energy_level
        pb_turn.steps_taken = agent.steps_taken

        pb_loc = PbLocation()
        pb_loc.x = agent.location.x
        pb_loc.y = agent.location.y
        pb_turn.loc.CopyFrom(pb_loc)

        pb_turn.spawns.extend(self.spawns)

        self.turns.append(pb_turn)
        self.clear_turn()

    def make_game_footer(self) -> None:
        if self.ws_server is None:
            error = "Server should have started."
            raise ValueError(error)

        game_footer = GameFooter()

        event = Event()
        event.game_footer.CopyFrom(game_footer)

        binary_string = event.SerializeToString()
        self.ws_server.add_event(binary_string)

    def make_games_footer(self) -> None:
        if self.ws_server is None:
            error = "Server should have started."
            raise ValueError(error)

        games_footer = GamesFooter()

        event = Event()
        event.games_footer.CopyFrom(games_footer)

        binary_string = event.SerializeToString()
        self.ws_server.add_event(binary_string)

    def add_team_info(self, team: Team, team_info: TeamInfo) -> None:
        pb_team_info = PbTeamInfo()
        pb_team_info.saved_alive = team_info.get_saved_alive(team)
        pb_team_info.saved_dead = team_info.get_saved_dead(team)
        pb_team_info.saved = team_info.get_saved(team)
        pb_team_info.predicted_right = team_info.get_predicted_right(team)
        pb_team_info.predicted_wrong = team_info.get_predicted_wrong(team)
        pb_team_info.predicted = team_info.get_predicted(team)
        pb_team_info.score = team_info.get_score(team)
        pb_team_info.units = team_info.get_units(team)
        pb_team_info.team = self.team_to_schema(team)
        self.team_info.append(pb_team_info)

    def add_spawn(self, agent_id: int, team: Team, loc: Location) -> None:
        pb_spawn = Spawn()
        pb_spawn.agentId = agent_id
        pb_loc = PbLocation()
        pb_loc.x = loc.x
        pb_loc.y = loc.y
        pb_spawn.loc.CopyFrom(pb_loc)
        pb_spawn.team = self.team_to_schema(team)
        self.spawns.append(pb_spawn)

    def add_removed_layer(self, loc: Location) -> None:
        pb_loc = PbLocation()
        pb_loc.x = loc.x
        pb_loc.y = loc.y
        self.removed_layers.append(pb_loc)

    def add_dead(self, agent_id: int) -> None:
        self.dead_ids.append(agent_id)

    def add_drone_scan(self, loc: Location, team: Team, duration: int) -> None:
        pb_drone_scan = DroneScan()
        pb_loc = PbLocation()
        pb_loc.x = loc.x
        pb_loc.y = loc.y
        pb_drone_scan.location.CopyFrom(pb_loc)
        pb_drone_scan.team = self.team_to_schema(team)
        pb_drone_scan.duration = duration
        self.drone_scans.append(pb_drone_scan)

    def add_survivor_health_update(
        self, location: Location, survivor_id: int, new_health: int, new_state: int
    ) -> None:
        """Add a survivor health update to be sent to the client."""
        pb_update = SurvivorHealthUpdate()
        pb_update.location.x = location.x
        pb_update.location.y = location.y
        pb_update.survivor_id = survivor_id
        pb_update.new_health = new_health
        pb_update.new_state = (
            SurvivorState.ALIVE if new_state == 0 else SurvivorState.DEAD
        )
        self.survivor_health_updates.append(pb_update)

    def team_to_schema(self, team: Team) -> PbTeam:
        return PbTeam.GOOBS if team == Team.GOOBS else PbTeam.VOIDSEERS

    def clear_round(self) -> None:
        """Clear all round data."""
        self.turns.clear()
        self.team_info.clear()
        self.removed_layers.clear()
        self.dead_ids.clear()
        self.drone_scans.clear()
        self.survivor_health_updates.clear()

    def clear_turn(self) -> None:
        self.spawns.clear()



import random
from typing import cast

import numpy as np
from numpy.typing import NDArray

from _aegis_game.args_parser import LaunchArgs
from _aegis_game.logger import LOGGER
from _aegis_game.team import Team
from _aegis_game.types.prediction import (
    CompletedPrediction,
    PendingPrediction,
)

from .data_loader import PredictionDataLoader

PendingPredictions = dict[tuple[Team, int], PendingPrediction]
CompletedPredictions = dict[tuple[Team, int], CompletedPrediction]


class PredictionHandler:
    def __init__(self, args: LaunchArgs) -> None:
        self._pending_predictions: PendingPredictions = {}
        self._completed_predictions: CompletedPredictions = {}
        self._data_loader: PredictionDataLoader = PredictionDataLoader(args)
        self._args: LaunchArgs = args

    def get_image_from_index(self, index: int) -> NDArray[np.uint8]:
        return cast("NDArray[np.uint8]", self._data_loader.x_test[index])

    def get_label_from_index(self, index: int) -> np.int32:
        return cast("np.int32", self._data_loader.y_test[index])

    def create_pending_prediction(self, team: Team, surv_id: int) -> None:
        """
        Create a pending prediction for a team-survivor combination.

        If one already exists, this method does nothing.
        """
        key: tuple[Team, int] = (team, surv_id)

        # Only create if no pending prediction exists
        if key not in self._pending_predictions:
            random_index = random.randint(0, len(self._data_loader.x_test) - 1)
            pending_prediction: PendingPrediction = {
                "image_to_predict": self.get_image_from_index(random_index),
                "correct_label": self.get_label_from_index(random_index),
            }
            self._pending_predictions[key] = pending_prediction

    def read_pending_predictions(
        self, team: Team
    ) -> list[tuple[int, NDArray[np.uint8], NDArray[np.int32]]]:
        """
        Agents call this to get all pending predictions for their team. Gives them the data of the pending prediction, without the correct label.

        Returns list of tuples: (surv_id, image_to_predict, all_unique_labels)
        """
        pending_list: list[tuple[int, NDArray[np.uint8], NDArray[np.int32]]] = []
        for (
            team_key,
            surv_id,
        ), pending_prediction in self._pending_predictions.items():
            if team_key == team:
                pending_list.append(
                    (
                        surv_id,
                        pending_prediction["image_to_predict"],
                        self._data_loader.unique_labels,
                    )
                )
        return pending_list

    def predict(self, team: Team, surv_id: int, prediction: np.int32) -> bool | None:
        """
        Process a prediction for a specific team-survivor combination.

        Returns:
            - bool: True if prediction was correct, False if incorrect
            - None: If no valid pending prediction exists (already predicted or never created)

        """
        key: tuple[Team, int] = (team, surv_id)

        # Check if there's a valid pending prediction
        if key not in self._pending_predictions:
            LOGGER.warning(
                f"Agent attempted to predict surv_id {surv_id} for team {team.name}, but no valid pending prediction exists. Did another agent on your team predict this survivor before you?"
            )
            return None

        pending_prediction = self._pending_predictions[key]
        is_correct: bool = pending_prediction["correct_label"] == prediction

        completed_prediction: CompletedPrediction = {
            "team": team,
            "surv_id": surv_id,
            "is_correct": is_correct,
        }

        self._completed_predictions[key] = completed_prediction
        del self._pending_predictions[key]

        return is_correct



from __future__ import annotations

from typing import override

from .world_object import WorldObject


class Rubble(WorldObject):
    """
    Represents a rubble object in the world.

    Attributes:
        id: Unique identifier for the rubble.
        energy_required: Amount of energy needed to remove the rubble.
        agents_required: Number of agents required to remove the rubble.

    """

    def __init__(
        self, rubble_id: int = -1, energy_required: int = 1, agents_required: int = 1
    ) -> None:
        super().__init__()
        self.id: int = rubble_id
        self.energy_required: int = energy_required
        self.agents_required: int = agents_required

    @override
    def __str__(self) -> str:
        return (
            f"RUBBLE ( ID {self.id} , "
            f"NUM_TO_RM {self.agents_required} , "
            f"RM_ENG {self.energy_required} )"
        )

    @override
    def __repr__(self) -> str:
        return self.__str__()



#!/usr/bin/env python3
"""
Script to fix line endings in generated protobuf files.
Converts CRLF (Windows) to LF (Unix) line endings.
"""

import glob
import os
from pathlib import Path


def fix_line_endings_in_file(file_path: str) -> int:
    """Convert CRLF line endings to LF in a single file.

    Returns:
        int: Number of CRLF sequences replaced
    """
    try:
        # Read file as binary to detect line endings
        with open(file_path, "rb") as f:
            content = f.read()

        # Check if file has CRLF endings
        if b"\r\n" in content:
            # Convert CRLF to LF
            new_content = content.replace(b"\r\n", b"\n")

            # Write back with LF endings
            with open(file_path, "wb") as f:
                f.write(new_content)

            replaced_count = content.count(b"\r\n")
            print(f"Fixed {replaced_count} CRLF sequences in: {file_path}")
            return replaced_count
        else:
            print(f"Already LF endings: {file_path}")
            return 0

    except Exception as e:
        print(f"Error processing {file_path}: {e}")
        return 0


def fix_line_endings_in_directory(directory: str, pattern: str = "*") -> int:
    """Fix line endings in all files matching the pattern in the directory.

    Returns:
        int: Total number of CRLF sequences replaced
    """
    directory_path = Path(directory)
    if not directory_path.exists():
        print(f"Directory does not exist: {directory}")
        return 0

    # Find all files matching the pattern
    search_pattern = directory_path / pattern
    files = glob.glob(str(search_pattern))

    if not files:
        print(f"No files found matching pattern: {search_pattern}")
        return 0

    print(f"Found {len(files)} files to process in {directory}")

    total_replaced = 0
    for file_path in files:
        if os.path.isfile(file_path):
            total_replaced += fix_line_endings_in_file(file_path)

    return total_replaced


def main():
    """Main function to fix line endings in all generated protobuf files."""
    print("Fixing line endings in generated protobuf files...")

    total_replaced = 0

    # Fix Python files
    print("\nProcessing Python files...")
    total_replaced += fix_line_endings_in_directory("python", "*.py")
    total_replaced += fix_line_endings_in_directory("python", "*.pyi")

    # Fix TypeScript files
    print("\nProcessing TypeScript files...")
    total_replaced += fix_line_endings_in_directory("ts", "*.ts")

    if total_replaced > 0:
        print(
            f"\nLine ending fix completed! Replaced {total_replaced} CRLF sequences with LF."
        )
    else:
        print("\nLine ending fix completed! All files already had LF endings.")


if __name__ == "__main__":
    main()



from collections.abc import Callable
from enum import Enum
from typing import Any, Literal, TypedDict, override


class CellType(Enum):
    """Enum representing different types of cells."""

    NORMAL_CELL = 0
    CHARGING_CELL = 1
    KILLER_CELL = 2
    SPAWN_CELL = 3

    @override
    def __str__(self) -> str:
        name = self.name.lower().replace("_cell", "")
        return name.capitalize()

    @override
    def __repr__(self) -> str:
        return self.__str__()


class GameOverReason(Enum):
    ALL_AGENTS_DEAD = "All team agents are dead"
    ALL_SURVIVORS_SAVED = "All survivors have been rescued"
    MAX_ROUNDS_REACHED = "Maximum number of rounds reached"


FeatureKey = Literal[
    "ALLOW_AGENT_PREDICTIONS",
    "ALLOW_AGENT_MESSAGES",
    "ALLOW_DRONE_SCAN",
    "ALLOW_AGENT_TYPES",
    "HIDDEN_MOVE_COSTS",
    "ALLOW_CUSTOM_AGENT_COUNT",
    "DEFAULT_AGENT_AMOUNT",
    "SURV_HEALTH_DECAY_RATE",
    "ADVANCED_SCORING_SYSTEM",
]

CompetitionSettingKey = Literal["VERSUS_MODE",]

ConfigType = Literal["assignment", "competition"]


class FeaturesConfig(TypedDict):
    ALLOW_AGENT_PREDICTIONS: bool
    ALLOW_AGENT_MESSAGES: bool
    ALLOW_DRONE_SCAN: bool
    ALLOW_AGENT_TYPES: bool
    HIDDEN_MOVE_COSTS: bool
    ALLOW_CUSTOM_AGENT_COUNT: bool
    DEFAULT_AGENT_AMOUNT: int
    SURV_HEALTH_DECAY_RATE: int
    ADVANCED_SCORING_SYSTEM: bool


class CompetitionConfig(TypedDict):
    VERSUS_MODE: bool


class ClientConfig(TypedDict):
    CONFIG_TYPE: ConfigType


class AegisConfig(TypedDict):
    features: FeaturesConfig
    competition_specific: CompetitionConfig
    client: ClientConfig


# `create_methods` return type
MethodDict = dict[str, type | Callable[..., Any]]  # pyright: ignore[reportExplicitAny]



import json
import argparse


def summarize_world(world_json):
    world = json.loads(world_json)

    # World info
    size = world["settings"]["world_info"]["size"]
    seed = world["settings"]["world_info"]["seed"]
    energy = world["settings"]["world_info"]["agent_energy"]
    print(f"World Size: {size['width']} x {size['height']}")
    print(f"Seed: {seed}")
    print(f"Agent Energy: {energy}\n")

    # Spawn locations (sorted)
    print("Spawn Locations:")
    for spawn in sorted(world["spawn_locs"], key=lambda s: (s["x"], s["y"])):
        print(f"  - ({spawn['x']}, {spawn['y']}) type: {spawn['type']}")
    print()

    # Killer cells (sorted)
    print("Killer Cells:")
    for kc in sorted(
        world["cell_types"].get("killer_cells", []), key=lambda c: (c["x"], c["y"])
    ):
        print(f"  - ({kc['x']}, {kc['y']})")
    print()

    # Charging cells (sorted)
    print("Charging Cells:")
    for cc in sorted(
        world["cell_types"].get("charging_cells", []), key=lambda c: (c["x"], c["y"])
    ):
        print(f"  - ({cc['x']}, {cc['y']})")
    print()

    # Stacks (sorted)
    print("Stacks with move_cost > 1 or contents:")
    for stack in sorted(
        world["stacks"], key=lambda s: (s["cell_loc"]["x"], s["cell_loc"]["y"])
    ):
        x, y = stack["cell_loc"]["x"], stack["cell_loc"]["y"]
        move_cost = stack["move_cost"]
        contents = stack.get("contents", [])
        if move_cost > 1 or contents:
            content_desc = (
                ", ".join([f"{c['type']}({c.get('arguments', {})})" for c in contents])
                or "none"
            )
            print(f"  - ({x},{y}) move_cost: {move_cost} contents: {content_desc}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Summarize a world JSON file.")
    parser.add_argument("file", type=str, help="Path to the world JSON file")
    args = parser.parse_args()

    with open(args.file, "r") as f:
        world_json = f.read()

    summarize_world(world_json)



from typing import override

from _aegis_game.types import CellType

from .location import Location
from .objects import WorldObject


class CellInfo:
    """
    Represents information about a cell in the world.

    Attributes:
        type: The type of the cell.
        location: The coordinates of the cell.
        move_cost: The movement cost to traverse this cell.
        agents: List of agent IDs currently in this cell.
        layers: Stack of world objects present in the cell.

    """

    def __init__(
        self,
        layers: list[WorldObject],
        cell_type: CellType,
        location: Location | None,
        move_cost: int,
        agents: list[int] | None,
    ) -> None:
        self.type: CellType = cell_type
        self.location: Location = location if location is not None else Location(-1, -1)
        self.move_cost: int = move_cost
        self.agents: list[int] = agents if agents is not None else []
        self.layers: list[WorldObject] = layers

    @property
    def top_layer(self) -> WorldObject | None:
        """Get the top-most layer of the cell."""
        return self.layers[0] if self.layers else None

    def is_killer_cell(self) -> bool:
        """
        Check if the cell is a killer cell.

        Returns:
            bool: True if the cell type is KILLER_CELL, else False.

        """
        return self.type == CellType.KILLER_CELL

    @override
    def __str__(self) -> str:
        return (
            f"{self.type.name} (\n"
            f"  X: {self.location.x},\n"
            f"  Y: {self.location.y},\n"
            f"  Move Cost: {self.move_cost},\n"
            f"  Num Agents: {len(self.agents)},\n"
            f"  Agent IDs: {self.agents},\n"
            f"  Top Layer: {self.top_layer}\n"
            f")"
        )

    @override
    def __repr__(self) -> str:
        return self.__str__()



from __future__ import annotations

from typing import override

from _aegis_game.types import CellType

from .location import Location
from .objects import Survivor, WorldObject


class Cell:
    def __init__(self, x: int, y: int) -> None:
        self.type: CellType = CellType.NORMAL_CELL
        self.layers: list[WorldObject] = []
        self.move_cost: int = 1
        self.agents: list[int] = []
        self.location: Location = Location(x, y)

    def setup_cell(self, cell_state_type: str) -> None:
        cell_state_type = cell_state_type.upper().strip()

        if cell_state_type == "CHARGING":
            self.type = CellType.CHARGING_CELL
        elif cell_state_type == "KILLER":
            self.type = CellType.KILLER_CELL
        elif cell_state_type == "SPAWN":
            self.type = CellType.SPAWN_CELL

    def is_charging_cell(self) -> bool:
        return self.type == CellType.CHARGING_CELL

    def is_killer_cell(self) -> bool:
        return self.type == CellType.KILLER_CELL

    def is_normal_cell(self) -> bool:
        return self.type == CellType.NORMAL_CELL

    def is_spawn_cell(self) -> bool:
        return self.type == CellType.SPAWN_CELL

    def set_spawn_cell(self) -> None:
        self.type = CellType.SPAWN_CELL

    def set_charging_cell(self) -> None:
        self.type = CellType.CHARGING_CELL

    def set_killer_cell(self) -> None:
        self.type = CellType.KILLER_CELL

    def get_layers(self) -> list[WorldObject]:
        return self.layers

    def add_layer(self, layer: WorldObject) -> None:
        self.layers.append(layer)

    def remove_top_layer(self) -> WorldObject:
        """
        Remove and returns the topmost layer from the cell.

        Assumes the caller has already checked that the cell has a top layer.

        Returns:
            WorldObject: The topmost world object layer.

        Raises:
            IndexError: If there are no layers to remove.

        """
        return self.layers.pop(0)

    def get_top_layer(self) -> WorldObject | None:
        if not self.layers:
            return None
        return self.layers[0]

    def set_top_layer(self, top_layer: WorldObject | None) -> None:
        self.layers.clear()
        if top_layer is None:
            return
        self.layers.append(top_layer)

    def number_of_survivors(self) -> int:
        count = 0
        for layer in self.layers:
            if isinstance(layer, Survivor):
                count += 1
        return count

    def is_spawn(self) -> bool:
        return self.type == CellType.SPAWN_CELL

    @override
    def __str__(self) -> str:
        if not self.layers:
            layers_str = "  (no layers)"
        else:
            layers_str = "\n".join(
                f"  {i + 1} - {layer}" for i, layer in enumerate(self.layers)
            )

        return (
            f"Cell at ({self.location.x}, {self.location.y}) - "
            f"Move Cost: {self.move_cost}\n"
            f"Type: {self.type}\n"
            f"Layers:\n"
            f"{layers_str}\n"
        )

    @override
    def __repr__(self) -> str:
        return self.__str__()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: team.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'team.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nteam.proto\x12\x05\x61\x65gis\"\xc0\x01\n\x08TeamInfo\x12\x13\n\x0bsaved_alive\x18\x01 \x01(\x05\x12\x12\n\nsaved_dead\x18\x02 \x01(\x05\x12\r\n\x05saved\x18\x03 \x01(\x05\x12\x17\n\x0fpredicted_right\x18\x04 \x01(\x05\x12\x17\n\x0fpredicted_wrong\x18\x05 \x01(\x05\x12\x11\n\tpredicted\x18\x06 \x01(\x05\x12\r\n\x05score\x18\x07 \x01(\x05\x12\r\n\x05units\x18\x08 \x01(\x05\x12\x19\n\x04team\x18\t \x01(\x0e\x32\x0b.aegis.Team* \n\x04Team\x12\t\n\x05GOOBS\x10\x00\x12\r\n\tVOIDSEERS\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'team_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TEAM']._serialized_start=216
  _globals['_TEAM']._serialized_end=248
  _globals['_TEAMINFO']._serialized_start=22
  _globals['_TEAMINFO']._serialized_end=214
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from enum import Enum


class Team(Enum):
    """
    Enum representing the team of an agent.

    Attributes:
        GOOBS (int): The Goobs team.
        VOIDSEERS (int): The Voidseers team.

    """

    GOOBS = 0
    """Team Goobs"""
    VOIDSEERS = 1
    """Team Voidseers"""

    def opponent(self) -> Team:
        """
        Get the opposing team.

        Returns:
            The opponent team of the current team.

        """
        return Team.VOIDSEERS if self == Team.GOOBS else Team.GOOBS



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Location(_message.Message):
    __slots__ = ("x", "y")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: int
    y: int
    def __init__(self, x: _Optional[int] = ..., y: _Optional[int] = ...) -> None: ...



# pyright: reportImportCycles = false
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

from .aegis_config import has_feature
from .agent import Agent
from .agent_type import AgentType
from .common import CellInfo, Direction, Location
from .common.objects.rubble import Rubble
from .common.objects.survivor import Survivor
from .constants import Constants
from .decorator import requires
from .message import Message
from .team import Team

if TYPE_CHECKING:
    from .game import Game


class AgentController:
    def __init__(self, game: "Game", agent: "Agent") -> None:
        self._game: Game = game
        self._agent: Agent = agent

    def assert_not_none(self, value: object) -> None:
        if value is None:
            error = "Argument has invalid None value"
            raise AgentError(error)

    def assert_spawn(self, loc: Location, team: Team) -> None:
        if loc not in self._game.get_spawns():
            error = f"Invalid spawn: {loc}"
            raise AgentError(error)

        units = self._game.team_info.get_units(team)
        if units == self._game.args.amount:
            error = "Max agents reached."
            raise AgentError(error)

    def assert_loc(self, loc: Location) -> None:
        self.assert_not_none(loc)
        if not self._game.on_map(loc):
            error = "Location is not on the map"
            raise AgentError(error)

    def assert_move(self, direction: Direction) -> None:
        self.assert_not_none(direction)
        new_loc = self._agent.location.add(direction)

        if not self._game.on_map(new_loc):
            error = "Agent moved off the map"
            raise AgentError(error)

    def assert_dig(self, agent: Agent) -> None:
        if has_feature("ALLOW_AGENT_TYPES") and agent.type not in (
            AgentType.ENGINEER,
            AgentType.COMMANDER,
        ):
            error = "Action not allowed. Only ENGINEER and COMMANDER can dig."
            raise AgentError(error)

    def assert_save(self, agent: Agent) -> None:
        if has_feature("ALLOW_AGENT_TYPES") and agent.type not in (
            AgentType.MEDIC,
            AgentType.COMMANDER,
        ):
            error = "Action not allowed. Only MEDIC and COMMANDER can save."
            raise AgentError(error)

    def assert_predict(self) -> None:
        if not has_feature("ALLOW_AGENT_PREDICTIONS"):
            msg = "Predictions are not enabled, therefore this method is not available."
            raise AgentError(msg)

    def assert_scan(self) -> None:
        if not has_feature("ALLOW_DRONE_SCAN"):
            msg = "Drone scan is not enabled, therefore this method is not available."
            raise AgentError(msg)

    # Public Agent Methods

    def get_round_number(self) -> int:
        """Return the current round number."""
        return self._game.round

    def get_id(self) -> int:
        """Return the id of the agent."""
        return self._agent.id

    def get_type(self) -> AgentType:
        """Return the type of the agent."""
        return self._agent.type

    def get_team(self) -> Team:
        """Return the current team of the agent."""
        return self._agent.team

    def get_location(self) -> Location:
        """Return the current location of the agent."""
        return self._agent.location

    def get_energy_level(self) -> int:
        """Return the current energy level of the agent."""
        return self._agent.energy_level

    def get_lumens(self) -> int:
        return self._game.team_info.get_lumens(self._agent.team)

    def move(self, direction: Direction) -> None:
        """
        Move the agent in the specified direction.

        Args:
            direction: The direction in which the agent should move.

        Raises:
            AgentError: If the move is invalid.

        """
        self.assert_move(direction)
        self._agent.add_cooldown()
        self._agent.apply_movement_cost(direction)
        new_loc = self._agent.location.add(direction)
        self._game.move_agent(self._agent.id, self._agent.location, new_loc)
        self._agent.location = new_loc

    def save(self) -> None:
        """
        Save a survivor located at the agent's current location.

        If no survivor is present, the function has no effect.

        Raises:
            AgentError: If saving is invalid according to game rules.

        """
        self.assert_save(self._agent)
        self._agent.add_cooldown()
        cell = self._game.get_cell_at(self._agent.location)
        top_layer = cell.get_top_layer()
        if top_layer is None or not isinstance(top_layer, Survivor):
            return

        self._game.save(top_layer, self._agent)

    def recharge(self) -> None:
        """
        Recharge the agent's energy if on a charging cell.

        Energy restored is equal to `Constants.NORMAL_CHARGE` per recharge,
        but cannot exceed `Constants.MAX_ENERGY_LEVEL`.

        Does nothing if the agent is not on a charging cell.
        """
        cell = self._game.get_cell_at(self._agent.location)
        if not cell.is_charging_cell():
            return

        energy = min(
            Constants.NORMAL_CHARGE,
            Constants.MAX_ENERGY_LEVEL - self._agent.energy_level,
        )

        self._agent.add_energy(energy)

    def dig(self) -> None:
        """
        Dig rubble at the agent's current location.

        Raises:
            AgentError: If digging is invalid according to game rules.

        """
        self.assert_dig(self._agent)
        self._agent.add_cooldown()
        cell = self._game.get_cell_at(self._agent.location)
        top_layer = cell.get_top_layer()
        if top_layer is None or not isinstance(top_layer, Rubble):
            return

        self._game.dig(self._agent)

    @requires("ALLOW_AGENT_PREDICTIONS")
    def predict(self, surv_id: int, label: np.int32) -> None:
        """
        Submit a prediction.

        Args:
            surv_id: The unique ID of the survivor.
            label: The predicted symbol label/classification.

        Raises:
            AgentError: If predictions are not enabled.

        """
        self.assert_predict()
        self._game.predict(surv_id, label, self._agent)

    @requires("ALLOW_AGENT_PREDICTIONS")
    def read_pending_predictions(
        self,
    ) -> list[tuple[int, NDArray[np.uint8], NDArray[np.int32]]]:
        """
        Retrieve the list of pending predictions stored by the agent's team.

        Each prediction is represented as a tuple containing:

            1. surv_id: The ID of the saved survivor that triggered this prediction.
            2. image_to_predict: The symbol image data for model input.
            3. all_unique_labels: The set of possible symbol labels.

        Returns:
            A list of pending symbol predictions. Returns an empty list if no pending
            predictions are available.

        Raises:
            AgentError: If predictions are not enabled.

        """
        self.assert_predict()
        return self._game.get_prediction_info_for_agent(self._agent.team)

    @requires("ALLOW_AGENT_MESSAGES")
    def send_message(self, message: str, dest_ids: list[int]) -> None:
        """
        Send a message to team members, excluding self.

        If `dest_ids` is empty, the message is broadcast to all team members
        except the sender.

        Args:
            message: The content of the message to send.
            dest_ids: A list of agent IDs to send the message to.

        """
        if not dest_ids:
            dest_ids = [
                agent.id
                for agent in self._game.agents.values()
                if agent.team == self._agent.team and agent.id != self._agent.id
            ]
        else:
            dest_ids = [aid for aid in dest_ids if aid != self._agent.id]

        msg = Message(
            message=message,
            round_num=self._game.round,
            sender_id=self._agent.id,
        )

        for agent_id in dest_ids:
            self._game.get_agent(agent_id).message_buffer.add_message(msg)

    @requires("ALLOW_AGENT_MESSAGES")
    def read_messages(self, round_num: int = -1) -> list[Message]:
        """
        Retrieve messages from the agent's message buffer.

        Args:
            round_num: If provided only messages from this round are returned.

        Returns:
            A list of messages.

        """
        if round_num == -1:
            return self._agent.message_buffer.get_all_messages()
        return self._agent.message_buffer.get_messages(round_num)

    @requires("ALLOW_DRONE_SCAN")
    def drone_scan(self, loc: Location) -> None:
        """
        Scan a location using a drone.

        Args:
            loc: The location to scan.

        Raises:
            AgentError: If drone scan is not enabled or location is invalid.

        """
        self.assert_scan()
        self.assert_loc(loc)

        self._game.start_drone_scan(loc, self._agent.team)
        self._agent.add_energy(-Constants.DRONE_SCAN_ENERGY_COST)

    def get_cell_info_at(self, loc: Location) -> CellInfo:
        """
        Return the cell info at a given location.

        If the location is adjacent (1 tile away) to the agent or has been scanned by a drone,
        all layers and agents at that location are visible. Otherwise, only the top layer is
        visible and agent presence is hidden.

        If `HIDDEN_MOVE_COSTS` feature is enabled, unvisited cells have `move_cost = 1`.

        Args:
            loc: The location to query.

        Returns:
            Information about the cell at the given location.

        """
        self.assert_loc(loc)

        idx = loc.x + loc.y * self._game.world.width
        cell_info = self._game.get_cell_info_at(loc)

        is_adjacent = self._agent.location.is_adjacent_to(loc)
        is_scanned = self._game.is_loc_drone_scanned(loc, self._agent.team)

        if not is_adjacent or not is_scanned:
            cell_info.agents = []
            top = cell_info.top_layer
            cell_info.layers = [top] if top is not None else []

        if has_feature("HIDDEN_MOVE_COSTS") and not self._agent.has_visited[idx]:
            cell_info.move_cost = 1

        return cell_info

    @requires("ALLOW_AGENT_TYPES")
    def spawn_agent(self, loc: Location, agent_type: AgentType) -> None:
        """
        Spawn an agent at a specified location with a given type.

        Args:
            loc: A valid spawn location.
            agent_type: The type of agent to spawn.

        Raises:
            AgentError: If spawn location is invalid or max amount reached.

        """
        self.assert_spawn(loc, self._agent.team)
        self._game.spawn_agent(loc, self._agent.team, agent_type)

    def log(self, *args: object) -> None:
        """
        Log a message.

        Args:
            *args: One or more items to log.

        """
        self._agent.log(*args)


class AgentError(Exception):
    """An error that occurs during agent actions or validations."""
